/******************************************************************************
 * $Id: coasp_dataset.cpp 13379 2007-12-19 02:39:14Z pvachon $
 *
 * Project:  DRDC Configurable Airborne SAR Processor (COASP) data reader
 * Purpose:  Support in GDAL for the DRDC COASP format data, both Metadata
 * 	     and complex imagery.
 * Author:   Philippe Vachon <philippe@cowpig.ca>
 * Notes:    I have seen a grand total of 2 COASP scenes (3 sets of headers).
 * 	     This is based on my best observations, some educated guesses and
 * 	     such. So if you have a scene that doesn't work, send it to me
 * 	     please and I will make it work... with violence.
 *
 ******************************************************************************
 * Copyright (c) 2007, Philippe Vachon
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 ****************************************************************************/

#include "gdal_priv.h"
#include "cpl_port.h"
#include "cpl_conv.h"
#include "cpl_vsi.h"
#include "cpl_string.h"

CPL_CVSID("$Id: coasp_dataset.cpp 13379 2007-12-19 02:39:14Z pvachon $");

CPL_C_START
void    GDALRegister_COASP(void);
CPL_C_END

#define TYPE_GENERIC 0
#define TYPE_GEOREF  1

enum ePolarization {
        hh = 0,
        hv,
        vh,
        vv
};


/*******************************************************************
 * Declaration of the COASPMetadata classes                        *
 *******************************************************************/

class COASPMetadataItem;

class COASPMetadataReader
{
        char **papszMetadata;
        int nMetadataCount;
        int nCurrentItem;
public:
        COASPMetadataReader(const char *pszFname);
        ~COASPMetadataReader();
        COASPMetadataItem *GetNextItem();
        COASPMetadataItem *GetItem(const char* pszName);
        int GotoMetadataItem(const char *pszName);
};

/* Your average metadata item */
class COASPMetadataItem
{
protected:
        char *pszItemName;
        char *pszItemValue;
public:
        COASPMetadataItem(const char *pszItemName = NULL, const char *pszItemValue = NULL);
        ~COASPMetadataItem();

        const char *GetItemName();
        const char *GetItemValue();
};

/* Same as MetadataItem class except parses GCP properly and returns
 * a GDAL_GCP struct 
 */
class COASPMetadataGeorefGridItem : public COASPMetadataItem
{
        int nId;
        int nPixels;
        int nLines;
        double ndLat;
        double ndLong;
public:
        COASPMetadataGeorefGridItem(int nId, int nPixels, int nLines, 
                double ndLat, double ndLong);
};

/********************************************************************
 * ================================================================ *
 * Implementation of the COASPMetadataItem Classes                  *
 * ================================================================ *
 ********************************************************************/

COASPMetadataItem::COASPMetadataItem(const char *pszItemName, const char *pszItemValue) 
{
    this->pszItemName = (pszItemName) ? VSIStrdup(pszItemName) : NULL;
    this->pszItemValue = (pszItemValue) ? VSIStrdup(pszItemValue) : NULL;
}

COASPMetadataItem::~COASPMetadataItem()
{
    CPLFree(pszItemName);
    CPLFree(pszItemValue);
}

const char *COASPMetadataItem::GetItemName() 
{
    return pszItemName;
}

const char *COASPMetadataItem::GetItemValue() 
{
    return pszItemValue;
}

COASPMetadataGeorefGridItem::COASPMetadataGeorefGridItem(int nId, int nPixels, 
	int nLines, double ndLat, double ndLong) : COASPMetadataItem("georef_grid", NULL)
{
    this->nId = nId;
    this->nPixels = nPixels;
    this->nLines = nLines;
    this->ndLat = ndLat;
    this->ndLong = ndLong;
}


/********************************************************************
 * ================================================================ *
 * Implementation of the COASPMetadataReader Class                  *
 * ================================================================ *
 ********************************************************************/

COASPMetadataReader::COASPMetadataReader(const char *pszFname) 
{
    papszMetadata = CSLLoad(pszFname);
    nMetadataCount = CSLCount(papszMetadata);
    nCurrentItem = 0;
}

COASPMetadataReader::~COASPMetadataReader()
{
    CSLDestroy(papszMetadata);
}

COASPMetadataItem *COASPMetadataReader::GetNextItem() 
{
    COASPMetadataItem *poMetadata;
    char **papszMDTokens;
    char *pszItemName;
    char *pszItemValue;

    if (nCurrentItem < 0 || nCurrentItem >= nMetadataCount)
        return NULL;

    papszMDTokens = CSLTokenizeString2(papszMetadata[nCurrentItem], " ", 
            CSLT_HONOURSTRINGS );
    pszItemName = papszMDTokens[0];
    if (pszItemName == NULL)
    {
        poMetadata = NULL;
    }
    else if (EQUALN(pszItemName, "georef_grid", 11))
    {
        if (CSLCount(papszMDTokens) >= 8)
        {
            double ndLat, ndLong;
            int nPixels, nLines;
            // georef_grid ( pixels lines ) ( lat long )
            // 0           1 2      3     4 5 6   7    8
            nPixels = atoi(papszMDTokens[2]);
            nLines = atoi(papszMDTokens[3]);
            ndLat = CPLAtof(papszMDTokens[6]);
            ndLong = CPLAtof(papszMDTokens[7]);
            poMetadata = new COASPMetadataGeorefGridItem(nCurrentItem, nPixels, 
                    nLines, ndLat, ndLong);
        }
        else
        {
            poMetadata = NULL;
        }
    }
    else
    {
        int nCount = CSLCount(papszMDTokens);
        if (nCount >= 2)
        {
            pszItemValue = CPLStrdup(papszMDTokens[1]); 
            for (int i = 2; i < nCount; i++)
            {
                int nSize = strlen(papszMDTokens[i]);
                pszItemValue = (char *)CPLRealloc(pszItemValue, 
                        strlen(pszItemValue) + 1 + nSize);
                strcat(pszItemValue, " ");
                strcat(pszItemValue, papszMDTokens[i]);
            }

            poMetadata = new COASPMetadataItem(pszItemName,
                    pszItemValue);

            CPLFree(pszItemValue);
        }
        else
        {
            poMetadata = NULL;
        }
    }
    CSLDestroy(papszMDTokens);
    nCurrentItem++;
    return poMetadata;
}

/* Goto the first metadata item with a particular name */
int COASPMetadataReader::GotoMetadataItem(const char *pszName)
{
    nCurrentItem = CSLPartialFindString(papszMetadata, pszName);
    return nCurrentItem;
}

COASPMetadataItem *COASPMetadataReader::GetItem(const char* pszName)
{
    GotoMetadataItem(pszName);
    return GetNextItem();
}


/*******************************************************************
 * Declaration of the COASPDataset class                           *
 *******************************************************************/


class COASPRasterBand;

/* A couple of observations based on the data I have available to me:
 * a) the headers don't really change, beyond indicating data sources
 *    and such. As such, I only read the first header specified by the 
 *    user. Note that this is agnostic: you can specify hh, vv, vh, hv and
 *    all the data needed will be immediately available.
 * b) Lots of GCPs are present in the headers. This is most excellent.
 * c) There is no documentation on this format. All the knowledge contained
 *    herein is from harassing various Defence Scientists at DRDC Ottawa.
 */

class COASPDataset : public GDALDataset
{
        friend class COASPRasterBand;

        int nGCPCount;
        GDAL_GCP *pasGCP;
public:
        COASPDataset();
        ~COASPDataset();

        static GDALDataset *Open( GDALOpenInfo * );
        static int Identify( GDALOpenInfo * poOpenInfo );
        int GetGCPCount();
        const GDAL_GCP *GetGCPs();
};


COASPDataset::COASPDataset()
{
    nGCPCount = 0;
    pasGCP = NULL;
}

COASPDataset::~COASPDataset()
{
    // TODO : free GCP when implemented
}

/********************************************************************
 * ================================================================ *
 * Declaration and implementation of the COASPRasterBand Class      *
 * ================================================================ *
 ********************************************************************/

class COASPRasterBand : public GDALRasterBand {
        FILE *fp;
        int ePol;
public:
        COASPRasterBand( COASPDataset *poDS, GDALDataType eDataType, int ePol, FILE *fp );
        ~COASPRasterBand();

        virtual CPLErr IReadBlock( int nBlockXOff, int nBlockYOff, 
                void *pImage);
};

COASPRasterBand::COASPRasterBand( COASPDataset *poDS, GDALDataType eDataType,
                                  int ePol, FILE *fp)
{
    this->fp = fp;
    this->ePol = ePol;
    this->poDS = poDS;
    this->eDataType = eDataType;
    this->nBlockXSize = poDS->GetRasterXSize();
    this->nBlockYSize = 1;
}

COASPRasterBand::~COASPRasterBand()
{
    if (fp)
        VSIFCloseL(fp);
}

CPLErr COASPRasterBand::IReadBlock( int nBlockXOff, int nBlockYOff, 
                                    void *pImage )
{
    /* 8 bytes per pixel: 4 bytes I, 4 bytes Q */
    unsigned long nByteNum = poDS->GetRasterXSize() * 8 * nBlockYOff;

    VSIFSeekL(fp, nByteNum, SEEK_SET);
    int nReadSize = (GDALGetDataTypeSize(eDataType)/8) * poDS->GetRasterXSize();
    VSIFReadL((char *)pImage, 1, nReadSize, fp);

#ifdef CPL_LSB
    GDALSwapWords( pImage, 4, nBlockXSize * 2, 4 ); 
#endif

    return CE_None;
}


/********************************************************************
 * ================================================================ *
 * Implementation of the COASPDataset Class                         *
 * ================================================================ *
 ********************************************************************/


/************************************************************************/
/*                            GetGCPCount()                             */
/************************************************************************/

int COASPDataset::GetGCPCount()
{
    return nGCPCount;
}

/************************************************************************/
/*                               GetGCPs()                              */
/************************************************************************/

const GDAL_GCP *COASPDataset::GetGCPs()
{
    return pasGCP;
}

/************************************************************************/
/*                              Identify()                              */
/************************************************************************/

int COASPDataset::Identify( GDALOpenInfo *poOpenInfo ) 
{
    if(poOpenInfo->fp == NULL || poOpenInfo->nHeaderBytes < 256)
            return 0;

    /* With a COASP .hdr file, the first line or so is:
        * time_first_datarec
        */
    if(EQUALN((char *)poOpenInfo->pabyHeader,"time_first_datarec",18))
            return 1;

    return 0;
}

/************************************************************************/
/*                                Open()                                */
/************************************************************************/

GDALDataset *COASPDataset::Open( GDALOpenInfo *poOpenInfo ) 
{
    if (!COASPDataset::Identify(poOpenInfo))
            return NULL;

    /* Set the binary matrix file pointers to NULL, for now */
    FILE* fpBinHH = NULL;
    FILE* fpBinHV = NULL;
    FILE* fpBinVH = NULL;
    FILE* fpBinVV = NULL;

    /* determine the file name prefix */
    char *pszBaseName = VSIStrdup(CPLGetBasename(poOpenInfo->pszFilename));
    char *pszDir = VSIStrdup(CPLGetPath(poOpenInfo->pszFilename));
    const char *pszExt = "rc";
    int nNull = strlen(pszBaseName) - 1;
    char *pszBase = (char *)CPLMalloc(nNull);
    strncpy(pszBase, pszBaseName, nNull);
    pszBase[nNull - 1] = '\0';
    CPLFree(pszBaseName);
    pszBaseName = NULL;

    char *psChan = strstr(pszBase,"hh");
    if (psChan == NULL) {
            psChan = strstr(pszBase, "hv");
    }
    if (psChan == NULL) {
            psChan = strstr(pszBase, "vh");
    }
    if (psChan == NULL) {
            psChan = strstr(pszBase, "vv");
    }

    if (psChan == NULL) {
            CPLFree(pszBase);
            CPLFree(pszDir);
            return NULL;
    }

    /* Create a fresh dataset for us to work with */
    COASPDataset *poDS;
    poDS = new COASPDataset();

    /* Read Metadata, set GCPs as is appropriate */
    COASPMetadataReader poReader(poOpenInfo->pszFilename);

    /* Get Image X and Y widths */
    COASPMetadataItem *poItem = poReader.GetItem("number_lines");
    if (!poItem)
    {
        CPLError(CE_Failure, CPLE_AppDefined, "Cannot read number_lines metadata.");
        delete poDS;
        CPLFree(pszBase);
        CPLFree(pszDir);
        return NULL;
    }

    poDS->nRasterYSize = atoi(poItem->GetItemValue());
    delete poItem;

    poItem = poReader.GetItem("number_samples");
    if (!poItem)
    {
        CPLError(CE_Failure, CPLE_AppDefined, "Cannot read number_samples metadata.");
        delete poDS;
        CPLFree(pszBase);
        CPLFree(pszDir);
        return NULL;
    }
    poDS->nRasterXSize = atoi(poItem->GetItemValue());
    delete poItem;

    const char* pszFilename;

    /* Horizontal transmit, horizontal receive */
    psChan[0] = 'h';
    psChan[1] = 'h';
    pszFilename = CPLFormFilename(pszDir, pszBase, pszExt); 

    fpBinHH = VSIFOpenL(pszFilename, "r");

    if (fpBinHH != NULL) {
        poDS->SetBand(1, new COASPRasterBand(poDS, GDT_CFloat32, 
                                             hh , fpBinHH));
    }

    /* Horizontal transmit, vertical receive */
    psChan[0] = 'h'; 
    psChan[1] = 'v'; 
    pszFilename = CPLFormFilename(pszDir, pszBase, pszExt);

    fpBinHV = VSIFOpenL(pszFilename, "r");

    if (fpBinHV != NULL) {
        poDS->SetBand(2, new COASPRasterBand(poDS, GDT_CFloat32,
                                             hv, fpBinHV));
    }

    /* Vertical transmit, horizontal receive */
    psChan[0] = 'v'; 
    psChan[1] = 'h'; 
    pszFilename = CPLFormFilename(pszDir, pszBase, pszExt);

    fpBinVH = VSIFOpenL(pszFilename, "r");

    if (fpBinVH != NULL) {
        poDS->SetBand(3, new COASPRasterBand(poDS, GDT_CFloat32,
                                             vh, fpBinVH));
    }

    /* Vertical transmit, vertical receive */
    psChan[0] = 'v'; 
    psChan[1] = 'v'; 
    pszFilename = CPLFormFilename(pszDir, pszBase, pszExt);

    fpBinVV = VSIFOpenL(pszFilename, "r");

    if (fpBinVV != NULL) {
        poDS->SetBand(4, new COASPRasterBand(poDS, GDT_CFloat32,
                                             vv, fpBinVV));
    }

    CPLFree(pszBase);
    CPLFree(pszDir);

    /* Oops, missing all the data? */

    if (fpBinHH == NULL && fpBinHV == NULL && fpBinVH == NULL && fpBinVV == NULL) 
    {
        CPLError(CE_Failure, CPLE_AppDefined, "Cannot find any data file.");
        delete poDS;
        return NULL;
    }

    if ( poDS->GetRasterCount() == 4 ) {
        poDS->SetMetadataItem( "MATRIX_REPRESENTATION", "SCATTERING" );
    }

    return poDS;
}

/************************************************************************/
/*                         GDALRegister_COASP()                         */
/************************************************************************/

void GDALRegister_COASP(void)
{
    GDALDriver *poDriver;
    if ( GDALGetDriverByName( "COASP" ) == NULL ) {
        poDriver = new GDALDriver();
        poDriver->SetDescription( "COASP" );
        poDriver->SetMetadataItem( GDAL_DMD_LONGNAME,
                "DRDC COASP SAR Processor Raster" );
        poDriver->SetMetadataItem( GDAL_DMD_EXTENSION,
                                   "hdr" );
/*      poDriver->SetMetadataItem( GDAL_DMD_HELPTOPIC,
                                   "frmt_coasp.html"); */
        poDriver->pfnIdentify = COASPDataset::Identify;
        poDriver->pfnOpen = COASPDataset::Open;
        GetGDALDriverManager()->RegisterDriver( poDriver );
    }

}
